""" Cli tool for testing connectivity with EQ3 smart thermostats. """
import logging
import re

import click

from eq3bt import Thermostat

pass_dev = click.make_pass_decorator(Thermostat)


def validate_mac(ctx, param, mac):
    if re.match("^([0-9A-Fa-f]{2}:){5}[0-9A-Fa-f]{2}$", mac) is None:
        raise click.BadParameter(mac + " is no valid mac address")
    return mac


@click.group(invoke_without_command=True)
@click.option("--mac", envvar="EQ3_MAC", required=True, callback=validate_mac)
@click.option("--interface", default=None)
@click.option("--debug/--normal", default=False)
@click.option(
    "--backend", type=click.Choice(["bleak", "bluepy", "gattlib"]), default="bleak"
)
@click.pass_context
def cli(ctx, mac, interface, debug, backend):
    """Tool to query and modify the state of EQ3 BT smart thermostat."""
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if backend == "bluepy":
        from .connection import BTLEConnection

        connection_cls = BTLEConnection
    elif backend == "gattlib":
        from .gattlibconnection import BTLEConnection

        connection_cls = BTLEConnection
    else:
        from .bleakconnection import BleakConnection

        connection_cls = BleakConnection

    thermostat = Thermostat(mac, interface, connection_cls)
    thermostat.update()
    ctx.obj = thermostat

    if ctx.invoked_subcommand is None:
        ctx.invoke(state)


@cli.command()
@click.option("--target", type=float, required=False)
@pass_dev
def temp(dev, target):
    """Gets or sets the target temperature."""
    click.echo("Current target temp: %s" % dev.target_temperature)
    if target:
        click.echo("Setting target temp: %s" % target)
        dev.target_temperature = target


@cli.command()
@click.option("--target", type=int, required=False)
@pass_dev
def mode(dev, target):
    """Gets or sets the active mode."""
    click.echo("Current mode: %s" % dev.mode_readable)
    if target:
        click.echo("Setting mode: %s" % target)
        dev.mode = target


@cli.command()
@click.option("--target", type=bool, required=False)
@pass_dev
def boost(dev, target):
    """Gets or sets the boost mode."""
    click.echo("Boost: %s" % dev.boost)
    if target is not None:
        click.echo("Setting boost: %s" % target)
        dev.boost = target


@cli.command()
@pass_dev
def valve_state(dev):
    """Gets the state of the valve."""
    click.echo("Valve: %s" % dev.valve_state)


@cli.command()
@click.option("--target", type=bool, required=False)
@pass_dev
def locked(dev, target):
    """Gets or sets the lock."""
    click.echo("Locked: %s" % dev.locked)
    if target is not None:
        click.echo("Setting lock: %s" % target)
        dev.locked = target


@cli.command()
@pass_dev
def low_battery(dev):
    """Gets the low battery status."""
    click.echo("Batter low: %s" % dev.low_battery)


@cli.command()
@click.option("--temp", type=float, required=False)
@click.option("--duration", type=float, required=False)
@pass_dev
def window_open(dev, temp, duration):
    """Gets and sets the window open settings."""
    click.echo("Window open: %s" % dev.window_open)
    if dev.window_open_temperature is not None:
        click.echo("Window open temp: %s" % dev.window_open_temperature)
    if dev.window_open_time is not None:
        click.echo("Window open time: %s" % dev.window_open_time)
    if temp and duration:
        click.echo(f"Setting window open conf, temp: {temp} duration: {duration}")
        dev.window_open_config(temp, duration)


@cli.command()
@click.option("--comfort", type=float, required=False)
@click.option("--eco", type=float, required=False)
@pass_dev
def presets(dev, comfort, eco):
    """Sets the preset temperatures for auto mode."""
    if dev.comfort_temperature is not None:
        click.echo("Current comfort temp: %s" % dev.comfort_temperature)
    if dev.eco_temperature is not None:
        click.echo("Current eco temp: %s" % dev.eco_temperature)
    if comfort and eco:
        click.echo(f"Setting presets: comfort {comfort}, eco {eco}")
        dev.temperature_presets(comfort, eco)


@cli.command()
@pass_dev
def schedule(dev):
    """Gets the schedule from the thermostat."""
    # TODO: expose setting the schedule somehow?
    for d in range(7):
        dev.query_schedule(d)
    for day in dev.schedule.values():
        click.echo(f"Day {day.day}, base temp: {day.base_temp}")
        current_hour = day.next_change_at
        for hour in day.hours:
            if current_hour == 0:
                continue
            click.echo(f"\t[{current_hour}-{hour.next_change_at}] {hour.target_temp}")
            current_hour = hour.next_change_at


@cli.command()
@click.argument("offset", type=float, required=False)
@pass_dev
def offset(dev, offset):
    """Sets the temperature offset [-3,5 3,5]"""
    if dev.temperature_offset is not None:
        click.echo("Current temp offset: %s" % dev.temperature_offset)
    if offset is not None:
        click.echo("Setting the offset to %s" % offset)
        dev.temperature_offset = offset


@cli.command()
@click.argument("away_end", type=click.DateTime(), default=None, required=False)
@click.argument("temperature", type=float, default=None, required=False)
@pass_dev
def away(dev, away_end, temperature):
    """Enables or disables the away mode."""
    if away_end:
        click.echo(f"Setting away until {away_end}, temperature: {temperature}")
    else:
        click.echo("Disabling away mode")
    dev.set_away(away_end, temperature)


@cli.command()
@pass_dev
def device(dev):
    """Displays basic device information."""
    dev.query_id()
    click.echo("Firmware version: %s" % dev.firmware_version)
    click.echo("Device serial:    %s" % dev.device_serial)


@cli.command()
@click.pass_context
def state(ctx):
    """Prints out all available information."""
    dev = ctx.obj
    click.echo(dev)
    ctx.forward(locked)
    ctx.forward(low_battery)
    ctx.forward(window_open)
    ctx.forward(boost)
    ctx.forward(temp)
    ctx.forward(presets)
    ctx.forward(offset)
    ctx.forward(mode)
    ctx.forward(valve_state)


if __name__ == "__main__":
    cli()



import codecs
from datetime import datetime, timedelta
from unittest import TestCase

import pytest

from eq3bt import TemperatureException, Thermostat
from eq3bt.eq3btsmart import PROP_ID_QUERY, PROP_INFO_QUERY, PROP_NTFY_HANDLE, Mode

ID_RESPONSE = b"01780000807581626163606067659e"
STATUS_RESPONSES = {
    "auto": b"020100000428",
    "manual": b"020101000428",
    "window": b"020110000428",
    "away": b"0201020004231d132e03",
    "boost": b"020104000428",
    "low_batt": b"020180000428",
    "valve_at_22": b"020100160428",
    "presets": b"020100000422000000001803282207",
}


class FakeConnection:
    def __init__(self, _iface, mac):
        self._callbacks = {}
        self._res = "auto"

    def set_callback(self, handle, cb):
        self._callbacks[handle] = cb

    def set_status(self, key):
        if key in STATUS_RESPONSES:
            self._res = key
        else:
            raise ValueError("Invalid key for status test response.")

    def make_request(self, handle, value, timeout=1, with_response=True):
        """Write a GATT Command without callback - not utf-8."""
        if with_response:
            cb = self._callbacks.get(PROP_NTFY_HANDLE)

            if value[0] == PROP_ID_QUERY:
                data = ID_RESPONSE
            elif value[0] == PROP_INFO_QUERY:
                data = STATUS_RESPONSES[self._res]
            else:
                return
            cb(codecs.decode(data, "hex"))


class TestThermostat(TestCase):
    def setUp(self):
        self.thermostat = Thermostat(
            _mac=None, _iface=None, connection_cls=FakeConnection
        )

    def test__verify_temperature(self):
        with self.assertRaises(TemperatureException):
            self.thermostat._verify_temperature(-1)
        with self.assertRaises(TemperatureException):
            self.thermostat._verify_temperature(35)

        self.thermostat._verify_temperature(8)
        self.thermostat._verify_temperature(25)

    @pytest.mark.skip()
    def test_parse_schedule(self):
        self.fail()

    @pytest.mark.skip()
    def test_handle_notification(self):
        self.fail()

    def test_query_id(self):
        self.thermostat.query_id()
        self.assertEqual(self.thermostat.firmware_version, 120)
        self.assertEqual(self.thermostat.device_serial, "PEQ2130075")

    def test_update(self):
        th = self.thermostat

        th._conn.set_status("auto")
        th.update()
        self.assertEqual(th.valve_state, 0)
        self.assertEqual(th.mode, Mode.Auto)
        self.assertEqual(th.target_temperature, 20.0)
        self.assertFalse(th.locked)
        self.assertFalse(th.low_battery)
        self.assertFalse(th.boost)
        self.assertFalse(th.window_open)

        th._conn.set_status("manual")
        th.update()
        self.assertTrue(th.mode, Mode.Manual)

        th._conn.set_status("away")
        th.update()
        self.assertEqual(th.mode, Mode.Away)
        self.assertEqual(th.target_temperature, 17.5)
        self.assertEqual(th.away_end, datetime(2019, 3, 29, 23, 00))

        th._conn.set_status("boost")
        th.update()
        self.assertTrue(th.boost)
        self.assertEqual(th.mode, Mode.Boost)

    def test_presets(self):
        th = self.thermostat
        self.thermostat._conn.set_status("presets")
        self.thermostat.update()
        self.assertEqual(th.window_open_temperature, 12.0)
        self.assertEqual(th.window_open_time, timedelta(minutes=15.0))
        self.assertEqual(th.comfort_temperature, 20.0)
        self.assertEqual(th.eco_temperature, 17.0)
        self.assertEqual(th.temperature_offset, 0)

    @pytest.mark.skip()
    def test_query_schedule(self):
        self.fail()

    @pytest.mark.skip()
    def test_schedule(self):
        self.fail()

    @pytest.mark.skip()
    def test_set_schedule(self):
        self.fail()

    @pytest.mark.skip()
    def test_target_temperature(self):
        self.fail()

    @pytest.mark.skip()
    def test_mode(self):
        self.fail()

    @pytest.mark.skip()
    def test_mode_readable(self):
        self.fail()

    @pytest.mark.skip()
    def test_boost(self):
        self.fail()

    def test_valve_state(self):
        th = self.thermostat
        th._conn.set_status("valve_at_22")
        th.update()
        self.assertEqual(th.valve_state, 22)

    def test_window_open(self):
        th = self.thermostat
        th._conn.set_status("window")
        th.update()
        self.assertTrue(th.window_open)

    @pytest.mark.skip()
    def test_window_open_config(self):
        self.fail()

    @pytest.mark.skip()
    def test_locked(self):
        self.fail()

    @pytest.mark.skip()
    def test_low_battery(self):
        th = self.thermostat
        th._conn.set_status("low_batt")
        th.update()
        self.assertTrue(th.low_battery)

    @pytest.mark.skip()
    def test_temperature_offset(self):
        self.fail()

    @pytest.mark.skip()
    def test_activate_comfort(self):
        self.fail()

    @pytest.mark.skip()
    def test_activate_eco(self):
        self.fail()

    @pytest.mark.skip()
    def test_min_temp(self):
        self.fail()

    @pytest.mark.skip()
    def test_max_temp(self):
        self.fail()

    @pytest.mark.skip()
    def test_away_end(self):
        self.fail()

    @pytest.mark.skip()
    def test_decode_mode(self):
        self.fail()



"""
Bleak connection backend.
This creates a new event loop that is used to integrate bleak's
asyncio functions to synchronous architecture of python-eq3bt.
"""
import asyncio
import codecs
import contextlib
import logging
from typing import Optional

from bleak import BleakClient, BleakError

from . import BackendException

DEFAULT_TIMEOUT = 1

# bleak backends are very loud on debug, this reduces the log spam when using --debug
logging.getLogger("bleak.backends").setLevel(logging.WARNING)

_LOGGER = logging.getLogger(__name__)


class BleakConnection:
    """Representation of a BTLE Connection."""

    def __init__(self, mac, iface):
        """Initialize the connection."""

        self._conn: Optional[BleakClient] = None
        self._mac = mac
        self._iface = iface
        self._callbacks = {}
        self._notifyevent = asyncio.Event()
        self._notification_handle = None
        self._loop = asyncio.new_event_loop()

    def __enter__(self):
        """
        Context manager __enter__ for connecting the device
        :rtype: BTLEConnection
        :return:
        """
        _LOGGER.debug("Trying to connect to %s", self._mac)

        kwargs = {}
        if self._iface is not None:
            kwargs["adapter"] = self._iface
        self._conn = BleakClient(self._mac, **kwargs)
        try:
            self._loop.run_until_complete(self._conn.connect())
        except BleakError as ex:
            _LOGGER.debug(
                "Unable to connect to the device %s, retrying: %s", self._mac, ex
            )
            try:
                self._loop.run_until_complete(self._conn.connect())
            except Exception as ex2:
                _LOGGER.debug("Second connection try to %s failed: %s", self._mac, ex2)
                raise BackendException(
                    "unable to connect to device using bleak"
                ) from ex2

        # The notification handles are off-by-one compared to gattlib and bluepy
        self._loop.run_until_complete(
            self._conn.start_notify(self._notification_handle - 1, self.on_notification)
        )
        _LOGGER.debug("Connected to %s", self._mac)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._conn:
            self._loop.run_until_complete(self._conn.disconnect())
            self._conn = None

    async def on_notification(self, handle, data):
        """Handle Callback from a Bluetooth (GATT) request."""
        # The notification handles are off-by-one compared to gattlib and bluepy
        handle = handle + 1
        _LOGGER.debug(
            "Got notification from %s: %s", handle, codecs.encode(data, "hex")
        )
        self._notifyevent.set()

        if handle in self._callbacks:
            self._callbacks[handle](data)

    @property
    def mac(self):
        """Return the MAC address of the connected device."""
        return self._mac

    def set_callback(self, handle, function):
        """Set the callback for a Notification handle. It will be called with the parameter data, which is binary."""
        self._notification_handle = handle
        self._callbacks[handle] = function

    async def wait_for_response(self, timeout):
        with contextlib.suppress(asyncio.TimeoutError):
            await asyncio.wait_for(self._notifyevent.wait(), timeout)

    def make_request(self, handle, value, timeout=DEFAULT_TIMEOUT, with_response=True):
        """Write a GATT Command without callback - not utf-8."""
        try:
            with self:
                _LOGGER.debug(
                    "Writing %s to %s",
                    codecs.encode(value, "hex"),
                    handle,
                )
                self._notifyevent.clear()

                self._loop.run_until_complete(
                    self._conn.write_gatt_char(handle - 1, value)
                )
                if timeout:
                    _LOGGER.debug("Waiting for notifications for %s", timeout)
                    self._loop.run_until_complete(self.wait_for_response(timeout))

        except BleakError as ex:
            _LOGGER.debug("Got exception from bleak while making a request: %s", ex)
            raise BackendException("Exception on write using bleak") from ex



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eq3bt', 'eq3bt.tests']

package_data = \
{'': ['*']}

install_requires = \
['bleak', 'click', 'construct']

extras_require = \
{'bluepy': ['bluepy>=1.0.5'], 'gattlib': ['gattlib']}

entry_points = \
{'console_scripts': ['eq3cli = eq3bt.eq3cli:cli']}

setup_kwargs = {
    'name': 'python-eq3bt',
    'version': '0.2',
    'description': 'EQ3 bluetooth thermostat support library',
    'long_description': '# python-eq3bt\n\nPython library and a command line tool for EQ3 Bluetooth smart thermostats, uses bleak (default), bluepy or gattlib for BTLE communication.\n\n# Features\n\n* Reading device status: locked, low battery, valve state, window open, target temperature, active mode\n* Writing settings: target temperature, auto mode presets, temperature offset\n* Setting the active mode: auto, manual, boost, away\n* Reading the device serial number and firmware version\n* Reading presets and temperature offset in more recent firmware versions.\n\n## Not (yet) supported\n\n* No easy-to-use interface for setting schedules.\n\n# Installation\n\n```bash\npip install python-eq3bt\n```\n\n# Command-line Usage\n\nTo test all available functionality a cli tool inside utils can be used:\n```\n$ eq3cli  --help\nUsage: eq3cli [OPTIONS] COMMAND [ARGS]...\n\n  Tool to query and modify the state of EQ3 BT smart thermostat.\n\nOptions:\n  --mac TEXT                  [required]\n  --interface TEXT\n  --debug / --normal\n  --backend [bleak|bluepy|gattlib]\n  --help                      Show this message and exit.\n\nCommands:\n  away         Enables or disables the away mode.\n  boost        Gets or sets the boost mode.\n  device       Displays basic device information.\n  locked       Gets or sets the lock.\n  low-battery  Gets the low battery status.\n  mode         Gets or sets the active mode.\n  offset       Sets the temperature offset [-3,5 3,5]\n  presets      Sets the preset temperatures for auto mode.\n  schedule     Gets the schedule from the thermostat.\n  state        Prints out all available information.\n  temp         Gets or sets the target temperature.\n  valve-state  Gets the state of the valve.\n  window-open  Gets and sets the window open settings.\n```\n\nEQ3_MAC environment variable can be used to define mac to avoid typing it:\n```bash\nexport EQ3_MAC=XX:XX\n```\n\nWithout parameters current state of the device is printed out.\n```bash\n$ eq3cli\n\n[00:1A:22:XX:XX:XX] Target 17.0 (mode: auto dst, away: no)\nLocked: False\nBatter low: False\nWindow open: False\nWindow open temp: 12.0\nWindow open time: 0:15:00\nBoost: False\nCurrent target temp: 17.0\nCurrent comfort temp: 20.0\nCurrent eco temp: 17.0\nCurrent mode: auto dst locked\nValve: 0\n```\n\nGetting & setting values.\n```bash\n$ eq3cli temp\n\nCurrent target temp: 17.0\n\neq3cli temp --target 20\n\nCurrent target temp: 17.0\nSetting target temp: 20.0\n```\n\n# Pairing\n\nIf you have thermostat with firmware version 1.20+ pairing may be needed. Below simple procedure to do that.\n\n```\nPress and hold wheel on thermostat until Pair will be displayed. Remember or write it.\n\n$ sudo bluetoothctl\n[bluetooth]# power on\n[bluetooth]# agent on\n[bluetooth]# default-agent\n[bluetooth]# scan on\n[bluetooth]# scan off\n[bluetooth]# pair 00:1A:22:06:A7:83\n[agent] Enter passkey (number in 0-999999): <enter pin>\n[bluetooth]# trust 00:1A:22:06:A7:83\n[bluetooth]# disconnect 00:1A:22:06:A7:83\n[bluetooth]# exit\n\nOptional steps:\n[bluetooth]# devices - to list all bluetooth devices\n[bluetooth]# info 00:1A:22:06:A7:83\nDevice 00:1A:22:06:A7:83 (public)\n\tName: CC-RT-BLE\n\tAlias: CC-RT-BLE\n\tPaired: yes\n\tTrusted: yes\n\tBlocked: no\n\tConnected: no\n\tLegacyPairing: no\n\tUUID: Generic Access Profile    (00001800-0000-1000-8000-00805f9b34fb)\n\tUUID: Generic Attribute Profile (00001801-0000-1000-8000-00805f9b34fb)\n\tUUID: Device Information        (0000180a-0000-1000-8000-00805f9b34fb)\n\tUUID: Vendor specific           (3e135142-654f-9090-134a-a6ff5bb77046)\n\tUUID: Vendor specific           (9e5d1e47-5c13-43a0-8635-82ad38a1386f)\n\tManufacturerData Key: 0x0000\n\tManufacturerData Value:\n  00 00 00 00 00 00 00 00 00                       .........\n```\n\nBe aware that sometimes if you pair your device then mobile application (calor BT) can\'t connect with thermostat and vice versa.\n\n\n# Library Usage\n\n```\nfrom eq3bt import Thermostat\n\nthermostat = Thermostat(\'AB:CD:EF:12:23:45\')\nthermostat.update()  # fetches data from the thermostat\n\nprint(thermostat)\n```\n\n<aside class="notice">\nNotice: The device in question has to be disconnected from bluetoothd, since BTLE devices can only hold one connection at a time.\n\nThe library will try to connect to the device second time in case it wasn\'t successful in the first time,\nwhich can happen if you are running other applications connecting to the same thermostat.\n</aside>\n\n## Fetching schedule\n\nThe schedule is queried per day basis and the cached information can be\naccessed through `schedule` property..\n\n```\nfrom eq3bt import Thermostat\n\nthermostat = Thermostat(\'AB:CD:EF:12:34:45\')\nthermostat.query_schedule(0)\nprint(thermostat.schedule)\n```\n\n## Setting schedule\n\nThe \'base_temp\' and \'next_change_at\' paramater define the first period for that \'day\' (the period from midnight up till next_change_at).\n\nThe schedule can be set on a per day basis like follows:\n\n```\nfrom datetime import time\nfrom eq3bt import Thermostat\nfrom eq3bt import HOUR_24_PLACEHOLDER as END_OF_DAY\n\nthermostat = Thermostat(\'12:34:56:78:9A:BC\')\nthermostat.set_schedule(\n    dict(\n        cmd="write",\n        day="sun",\n        base_temp=18,\n        next_change_at=time(8, 0),\n        hours=[\n            dict(target_temp=23, next_change_at=time(20, 0)),\n            dict(target_temp=18, next_change_at=END_OF_DAY),\n            dict(target_temp=23, next_change_at=END_OF_DAY),\n            dict(target_temp=23, next_change_at=END_OF_DAY),\n            dict(target_temp=23, next_change_at=END_OF_DAY),\n            dict(target_temp=23, next_change_at=END_OF_DAY)\n        ]\n    )\n)\n```\n\n# Contributing\n\nFeel free to open pull requests to improve the library!\n\nThis project uses github actions to enforce code formatting using tools like black, isort, flake8, and mypy.\nYou can run these checks locally either by executing `pre-commit run -a` or using `tox` which also runs the test suite.\n\n\n# History\n\nThis library is a simplified version of bluepy_devices from Markus Peter (https://github.com/bimbar/bluepy_devices.git) with support for more features and robuster device handling.\n',
    'author': 'Teemu R.',
    'author_email': 'tpr@iki.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rytilahti/python-eq3bt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
A simple wrapper for bluepy's btle.Connection.
Handles Connection duties (reconnecting etc.) transparently.
"""
import codecs
import logging

from bluepy import btle

from . import BackendException

DEFAULT_TIMEOUT = 1

_LOGGER = logging.getLogger(__name__)


class BTLEConnection(btle.DefaultDelegate):
    """Representation of a BTLE Connection."""

    def __init__(self, mac, iface):
        """Initialize the connection."""
        btle.DefaultDelegate.__init__(self)

        self._conn = None
        self._mac = mac
        self._iface = iface
        self._callbacks = {}

    def __enter__(self):
        """
        Context manager __enter__ for connecting the device
        :rtype: btle.Peripheral
        :return:
        """
        self._conn = btle.Peripheral()
        self._conn.withDelegate(self)
        _LOGGER.debug("Trying to connect to %s", self._mac)
        try:
            self._conn.connect(self._mac, iface=self._iface)
        except btle.BTLEException as ex:
            _LOGGER.debug(
                "Unable to connect to the device %s, retrying: %s", self._mac, ex
            )
            try:
                self._conn.connect(self._mac, iface=self._iface)
            except Exception as ex2:
                _LOGGER.debug("Second connection try to %s failed: %s", self._mac, ex2)
                raise BackendException(
                    "Unable to connect to device using bluepy"
                ) from ex2

        _LOGGER.debug("Connected to %s", self._mac)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._conn:
            self._conn.disconnect()
            self._conn = None

    def handleNotification(self, handle, data):
        """Handle Callback from a Bluetooth (GATT) request."""
        _LOGGER.debug(
            "Got notification from %s: %s", handle, codecs.encode(data, "hex")
        )
        if handle in self._callbacks:
            self._callbacks[handle](data)

    @property
    def mac(self):
        """Return the MAC address of the connected device."""
        return self._mac

    def set_callback(self, handle, function):
        """Set the callback for a Notification handle. It will be called with the parameter data, which is binary."""
        self._callbacks[handle] = function

    def make_request(self, handle, value, timeout=DEFAULT_TIMEOUT, with_response=True):
        """Write a GATT Command without callback - not utf-8."""
        try:
            with self:
                _LOGGER.debug(
                    "Writing %s to %s with with_response=%s",
                    codecs.encode(value, "hex"),
                    handle,
                    with_response,
                )
                self._conn.writeCharacteristic(
                    handle, value, withResponse=with_response
                )
                if timeout:
                    _LOGGER.debug("Waiting for notifications for %s", timeout)
                    self._conn.waitForNotifications(timeout)
        except btle.BTLEException as ex:
            _LOGGER.debug("Got exception from bluepy while making a request: %s", ex)
            raise BackendException("Exception on write using bluepy") from ex
